#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <signal.h>
#include <setjmp.h>
#include <time.h>

#include "tetris.h"
#include "io.h"

#include "misc.h"
#include "music.h"

#include "music_cv1d.h"
#include "music_cv2b.h"
#include "music_tet1.h"
#include "music_tet4.h"
#include "music_tetb.h"
#include "music_tetk.h"
#include "music_tetl.h"
#include "music_tett.h"
#include "music_teta.h"
#include "music_tetr.h"
#include "music_teto.h"
#include "music_sols.h"
#include "music_none.h"

static jmp_buf ResetGame;
static tetris_t Game, Game2;

static void DrawMainBorder(void)
{
	int x, y;
	for(y=0; y<24; y++)
	{
		SetChar( 0,y, corner(c_up+c_dn+c_rt),mainborder, 2,0);
		SetChar( 1,y, corner(c_up+c_dn+c_lt),mainborder, 2,0);
		SetChar(38,y, corner(c_up+c_dn+c_rt),mainborder, 2,0);
		SetChar(39,y, corner(c_up+c_dn+c_lt),mainborder, 2,0);
	}
	for(x=1; x<39; x++)
	{
		SetChar(x, 0, corner(c_lt+c_rt),mainborder, 2,0);
		SetChar(x,23, corner(c_lt+c_rt),mainborder, 2,0);
	}
}

static void DrawScoreBorder(void)
{
	int x, y;
	
	for(x=20-9; x<20+9; x++)
		SetChar(x,4, corner(c_lt+c_rt), scoreborder,1,0);
			
	for(y=1; y<=4; y++)
	{
		SetChar(20-9,y, corner(c_up+(y<4?c_dn:c_rt)), scoreborder,1,0);
		SetChar(20+9,y, corner(c_up+(y<4?c_dn:c_lt)), scoreborder,1,0);
		SetChar(20-3,y, corner(c_up+(y<4?c_dn:c_lt+c_rt)), scoreborder,1,0);
		SetChar(20+3,y, corner(c_up+(y<4?c_dn:c_lt+c_rt)), scoreborder,1,0);
	}
	
	Center(1, scorecolor,0,0, "Score\nLines\nLevel");
}

static void GameBorder(void)
{
	int y, x;
	
	Center(20, 7,2,0, "%31s", "");
	Center(21, 7,2,0, "%27s", "");
	
	for(x=0; x<rakepax; x++)
	{
		SetChar(27-x,23, corner(c_lt+c_rt+c_up), mainborder,2,0);
		SetChar(12+x,23, corner(c_lt+c_rt+c_up), mainborder,2,0);
	}
	for(y=5+x; y<23; y++)
		for(x=0; x<rakepax; x++)
		{
			SetChar(12+x,y, corner(c_up+c_dn), rakennelma,2,0);
			SetChar(27-x,y, corner(c_up+c_dn), rakennelma,2,0);
		}
	for(y=0; y<rakepax; y++)
		for(x=12; x<=27; x++)
		{
			int c = (x<12+y||x>27-y)
				?	c_up+c_dn
				:	(x>12+y?c_lt:c_dn) + (x<27-y?c_rt:c_dn);
			SetChar(x,5+y, corner(c), rakennelma,2,0);
		}
}

static int Tick(int numticks)
{
	UpdateVideo(numticks);
	TickMusa(numticks);
	return numticks;
}

static int WhichKey(void)
{
	int c = getch();
	if(c=='A'||c=='w'||c=='H')c='';
	if(c=='B'||c=='s'||c=='P')c='';
	if(c=='D'||c=='a'||c=='K')c='';
	if(c=='C'||c=='d'||c=='M')c='';
	if(c=='\r')c='\n';
	return c;
}

static void MusicReactor(int MusicNum)
{
	switch(MusicNum)
	{
		case -2:StartMusa(tet1); break; /* main */
		case -1:break;					/* enter key */
		case 0:	StartMusa(none); break;
		case 1: StartMusa(tet4); break;
		case 2: StartMusa(cv1d); break;
		case 3: StartMusa(cv2b); break;
		case 4: StartMusa(tetl); break;
		case 5: StartMusa(tetb); break;
		case 6: StartMusa(tetk); break;
		case 7: StartMusa(tett); break;
		case 8: StartMusa(teta); break;
		case 9: StartMusa(tetr); break;
		case 10:StartMusa(sols); break;
	}
}

static computer_t *GetComp(int compnro)
{
	switch(compnro)
	{
		case 0: return &comp_dip;
		case 1: return &comp_dat;
		case 2:	return &comp_blip;
		case 3: return &comp_dak;
		case 4: return &comp_sim;
	}
	return NULL;
}

static void CompReactor(int CompNum)
{
	computer_t *tmp = GetComp(CompNum);

	Center(19, menucolor,2,0, "%36s\n%36s\n%36s","","","");
	if(tmp)Center(19, menucolor,2,0, tmp->Description);
}

BEGIN_EFFECT(eff_move, CHANNEL 0, TEMPO 400, LENGTH 4)
	50, 50,
	54, 50,
	60, 50,
	74, 50
END_EFFECT;

BEGIN_EFFECT(eff_sel, CHANNEL 0, TEMPO 300, LENGTH 6)
	50, 50,
	47, 45,
	44, 40,
	41, 35,
	38, 30,
	35, 25
END_EFFECT;

/* Reactor args:
 *      -2=quit
 *      -1=affirmated
 *   other=selection no
 */
static int Menu(void (*Reactor)(int), const char *Title, ...)
{
	int a, SelCount, Cur, Longest, x, Ticks, top=11;
	va_list ap;
	
	va_start(ap, Title);
	for(Longest=0, x=0;;)
	{
		char *s = va_arg(ap, char *);
		int Len;
		
		if(x++ > 8)top--;
		
		if(!s)break;
		Len=strlen(s);
		if(Len>Longest)Longest=Len;
	}
	va_end(ap);
	
	/* Menu is transparent - good? */
	Center(top-2, menucolor,1,0, Title);
	
	x = 21 - Longest/2;
	
	va_start(ap, Title);
	for(SelCount=0;;)
	{
		char *s = va_arg(ap, char *);
		if(!s)break;
		
		SetText(x, top+SelCount++, menucolor,1,0, "%s", s);
	}
	va_end(ap);
	
	Cur=0;
	
Redraw:
	if(Reactor)Reactor(Cur);
	for(a=0; a<SelCount; a++)
	{
		SetChar(x-3, top+a, a==Cur?'-':' ', menuarrow,1,0);
		SetChar(x-2, top+a, a==Cur?'>':' ', menuarrow,1,0);
	}
	for(;;)
	{
		Ticks=0;
		while(!kbhit())
		{
			Tick(1);
			if(++Ticks >= 10 SECONDS)
				longjmp(ResetGame, 1);
		}
		switch(WhichKey())
		{
			case '':	
				Cur=(Cur+SelCount-1)%SelCount;
				PlayEff(eff_move);
				goto Redraw;
			case '':
				Cur=(Cur+         1)%SelCount;
				PlayEff(eff_move);
				goto Redraw;
			case '':
			case '':
				PlayEff(eff_move);
				break;
			case 'Q':
			case 'q':
				return -2;
			case '\n':
				PlayEff(eff_sel);
				if(Reactor)Reactor(-1);
				for(a=-2; a<SelCount; a++)
					SetText(2,top+a, 7,1,0, "%36s", "");
				return Cur;
		}
	}
}

static void CtrlC(int dummy)
{
	DoneVideo();
	exit(dummy);
}

/* SEGV/FPE handler */
static void Problem(int dummy)
{
	dummy=dummy;
	DoneVideo();
}

static int GameMode;
static int Handicap;
static int Level;
static int Music;
	
#define KEYSETUPCOUNT 5
static int KeyMode;

static const char *KeySetups[KEYSETUPCOUNT] =
{
	"rotate1,drop,stepdown",
	"rotate1,rotate2,stepdown",
	"rotate2,rotate1,drop",
	"rotate2,rotate1,stepdown",
	"rotate2,drop,stepdown"
};

static const int Key[KEYSETUPCOUNT][3] =
{	
	{KEY_TURN1, KEY_DROP, KEY_DOWN},
	{KEY_TURN1, KEY_TURN2, KEY_DOWN},
	{KEY_TURN2, KEY_TURN1, KEY_DROP},
	{KEY_TURN2, KEY_TURN1, KEY_DOWN},
	{KEY_TURN2, KEY_DROP, KEY_DOWN}
};	
	
/* Return values:
 *   1=Play once more (start from the menus)
 *   0=Do not play, end the program. (Never used)
 */
static int GameLoop(void)
{
	MusicReactor(Music);
	
	for(;;)
	{
		time_t Time   = time(NULL);
		struct tm *TM = localtime(&Time);
		char md[40];
		
		strftime(md, sizeof(md)-1, "%A\n %d.%m.%Y ", TM);
		Center(19, clockcolor,1,0, "%9s", "");
		Center(19, clockcolor,1,0, md);
		Center(21, clockcolor,1,0,
			"%2d:%02d:%02d", TM->tm_hour, TM->tm_min, TM->tm_sec);
		
		switch(GameMode)
		{
			case 1:
				if(Game2.State == S_GAME)
				{
					if(Game2.Score < Game.Score
					&& Game2.CompSpeed > 1)
						Game2.CompSpeed--;
					if(Game2.Score > Game.Score
					&& Game2.CompSpeed < 0.5 SECONDS)
						Game2.CompSpeed++;
				}
				TickTetris(&Game2, 1);
			case 0:
				TickTetris(&Game, 1);
				break;
		}
		
		while(kbhit())
		{
			int a = WhichKey();
			
			switch(a)
			{
				case '':
					a = KEY_LT;
					break;
				case '':
					a = KEY_RT;
					break;
				case '':
					RefreshScreen();
					break;
				case '\n':
				case 'p':
					if(Game.State == S_OVER) return 1;	/* game over */
					a = KEY_PAUSE;
					break;
				case '':
					a = Key[KeyMode][0];
					break;
				case ' ':
					a = Key[KeyMode][1];
					break;
				case '':
					a = Key[KeyMode][2];
					break;
				case 'Q':
				case 'q':
					return 1; /* escape */
				default:
					/* Do not confuse the game */
					a=0;
					break;
			}
			
			if(a)
			{
				if(Game.Computer == Human)GrantKey(&Game, a);
				if(Game2.Computer == Human)GrantKey(&Game2, a);
			}
		}
		
		Tick(1);
		
		if(Game.State == S_OVER)
			Game2.NoHumanOpponents=1;
	}
}

static int SelectIO(const char *drv)
{
	extern iodriver ggi;
	extern iodriver vt100;
	InstallDriver(&ggi, &vt100, NULL);
	return SelectDriver(drv);
}

int main(int argc, const char **argv)
{
	int Comp1, Comp2;
	const char *drv = DEFAULT_IO;
	
	for(Comp1=0; --argc; )
	{
		extern int xres,yres, vxres,vyres;
		const char *s = *++argv;
		if(*s=='-')
		{
		Usage:
			printf(
				"Usage: joeltris10 [ggi | vt100] [xres yres vxres vyres]\n"
				"Default io driver: %s\n"
				"Default resolution: auto auto  640 384\n"
				"Read libggi documentation for help on using ggi\n",
					DEFAULT_IO);
			return 0;
		}
		switch(Comp1)
		{
			case 0: drv=s; break;
			case 1: xres=atoi(s); break;
			case 2: yres=atoi(s); break;
			case 3: vxres=atoi(s); break;
			case 4: vyres=atoi(s); break;
		}
		Comp1++;
	}
	
	if(SelectIO(drv) < 0)
	{
		printf("No driver named '%s' found\n", drv);
		goto Usage;
	}
	
	signal(SIGINT, CtrlC);
	#ifdef SIGPIPE
	signal(SIGPIPE, Problem);
	#endif
	#ifdef SIGFPE
	signal(SIGFPE, Problem);
	#endif
	signal(SIGSEGV, Problem);
	
	Randomize();
	
	InitVideo();
	
	if(setjmp(ResetGame))
		SetText(0,0, 8,1,2 SECONDS, "reset");
	
	ClrScr();
	
	MusicReactor(-2);
	
	Center(1, 6,1,2.2 SECONDS,
		"Playing this game with square font\n"
		"(8x8 etc) is strongly recommended");
	
	Center(9,  4,1,0.5 SECONDS, VERSION);
	
	Center(20, 1,1,2.5 SECONDS, "Written by Bisqwit");
	Center(21, 1,1,2.5 SECONDS, "http://iki.fi/bisqwit/");
	
	Center(8, 12,1,0, "JOELTRIS 10");
	Center(14,12,1,0, "PRESS ENTER");
	
	if(1)	/* Just for the subvars */
	{
	#if starcount
		int sx[starcount], sy[starcount], sd[starcount], sp[starcount], a;
		for(a=starcount; a--; sd[a]=0)
		{
			sx[a]=rand()%40;
			sy[a]=rand()%24;
			sp[a]=rand()&3;
			if(!sp[a])sp[a]=rand()&3;
		}	
	#endif
	#if enabletetris
		comp_dip.UseStackHeight=0;
		/*		 level,handicap, x,y */
		InitTetris(&Game,  0,0,  0,4,  -1,-1, 24,36, &comp_dip);
		Game.NoHumanOpponents=1;
	//	InitTetris(&Game2, 0,0, 12,4,  -1,-1, 12,36, &comp_dip);
	//	Game2.NoHumanOpponents=1;
	#endif
		
		for(;;)
		{
			while(!kbhit())
			{
			#if enabletetris
				TickTetris(&Game, 1); Game.Lines=0;
			//	TickTetris(&Game2, 1); Game2.Lines=0;
			#endif
			#if starcount
				for(a=0; a<starcount; a++)
					#if !enabletetris
					if(sd[a])sd[a]--;
					else
					#endif
					{
						static const int C[] = {15,7,7,8};
						register int c;
						
						#define cheka(tet) \
							if(tet.Field[y][x]!=' ')ok=0; \
							if(ok && tet.State>=S_GAME \
								&& x>=tet.blockx && x<=tet.blockx+4 \
								&& y>=tet.blocky && y<=tet.blocky+4) \
								ok = tet.ThisBlock[y-tet.blocky][x-tet.blockx]==' ';
						
					#if 0
						#define checkok \
						{ \
							int y=39-sx[a], x=sy[a]; ok=1; \
							if(x < 12) { cheka(Game); } \
							else { x-=12; cheka(Game2); } \
						}
					#else
						#define checkok \
						{ \
							int y=39-sx[a], x=sy[a]; ok=1; \
							cheka(Game); \
						}
					#endif
						
						#if enabletetris
						register int ok;
						
						checkok;
						if(ok)
						#endif
							SetChar(sx[a],sy[a], ' ',7,0,0);
							
						#if enabletetris
						if(sd[a])
							sd[a]--;
						else
						{
						#endif
							/* Timer */
							sd[a]=sp[a];
							if(--sx[a] < 0)sx[a]=39, sy[a]=rand()%24;
						#if enabletetris
						}
						
						checkok;
						if(ok)
						{
						#endif
							c = sp[a]>2?'.':'-';
							SetChar(sx[a],sy[a], c,C[sp[a]],0,0);
						#if enabletetris
						}
						#endif
					}
			#endif
				Tick(1);
			}
			if(WhichKey()=='\n')break;
		}
		#if enabletetris
		KillTetris(&Game);
//		KillTetris(&Game2);
		comp_dip.UseStackHeight=1;
		#endif
	}
	
Over:
	Comp1 = Comp2 = -1;
	
	ClrScr(); /* Wipe out the stars */
	DrawMainBorder();
	
	Center(8,  7,1,0, "%11s", "");
	Center(14, 7,1,0, "%11s", "");
	
	DrawScoreBorder();

	StartSilence();	/* Stop any music */
	
	GameMode = Menu(
		NULL,
		"GAME SELECT",
		"1 player",
		"versus computer",
		"intro",
		NULL);
		
	if(GameMode==-2)goto End;
	if(GameMode==2)longjmp(ResetGame, 1);	
	if(GameMode==1)
	{
		Comp2 = Menu(
			CompReactor,
			"COMPUTER OPPONENT SELECT",
			GetComp(0)->Name,
			GetComp(1)->Name,
			GetComp(2)->Name,
			GetComp(3)->Name,
			GetComp(4)->Name,
			NULL);
		if(Comp2==-2)goto Over;
	}
	
	KeyMode = Menu(
		NULL,
		"KEYBOARD BINDINGS (UP, SPACE & DOWN)",
		KeySetups[0],
		KeySetups[1],
		KeySetups[2],
		KeySetups[3],
		KeySetups[4],
		NULL);
	if(KeyMode==-2)goto Over;
	
	Level = Menu(
		NULL,
		"LEVEL SELECT",
		"0", "1", "2", "3", "4", "5", "6", "7", "8", "9",
		NULL);
	if(Level==-2)goto Over;
		
	Handicap = Menu(
		NULL,
		"HANDICAP SELECT",
		"0", "3", "6", "9", "12",
		NULL) * 3;
	if(Handicap==-2)goto Over;
		
	Center(20, 1,1,3 SECONDS, "Some of the music melodies are");
	Center(21, 1,1,3 SECONDS, "(C) Nintendo and (C) Konami");

	Music = Menu(
		MusicReactor,
		"MUSIC SELECT",
		"Silence",
		"Dance",
		"Frog",
		"Mansion",
		"Loginska",
		"Bradinsky",
		"Karinka",
		"Troika",
		"ASO Tetris theme",
		"ASO Technotris",
		"Solstice",
		NULL);
	if(Music==-2)goto Over;

	GameBorder();
	
	switch(GameMode)
	{
		case 0:	/* 1 player */
			InitTetris(&Game,  Level,Handicap,  2,5,  2,12, 10,18, NULL);
			break;
		case 1:	/* versus computer */
			InitTetris(&Game,  Level,Handicap,  2,5,  2,12, 10,18, NULL);
			InitTetris(&Game2, Level,Handicap, 28,5, 34,24, 10,18, GetComp(Comp2));
			Game2.MayMoan=1;
			break;
	}
	
	if(GameLoop())goto Over;
End:
	
	KillTetris(&Game);
	KillTetris(&Game2);
	
	DoneMusa();
	DoneVideo();
	
	return 0;
}
